#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
File containing definitions for the capability version routes of the Workspaces REST API
"""

import requests
from pycapo import CapoConfig
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPNotFound,
    HTTPPreconditionFailed,
)
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config


@view_config(route_name="view_all_versions", renderer="json")
def view_all_versions(request: Request):
    """
    Pyramid view that accepts a request to view all versions of a capability request
    URL: capability/request/{capability_request_id}/versions

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.versions:
            return capability_request.versions
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_latest_version", renderer="json")
def view_latest_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.current_version:
            version_json = capability_request.current_version.__json__()
            version_json["files"] = []

            for file in capability_request.current_version.files:
                version_json["files"].append(file.__json__())

            return Response(json_body=version_json)
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_specific_version", renderer="json")
def view_specific_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            return version
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="create_capability_version", renderer="json")
def create_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a capability version
    URL: capability/request/{capability_request_id}/version/create

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """
    expected_params = ["parameters"]
    capability_request_id = request.matchdict["capability_request_id"]
    params = request.json_body

    if not all([expected in params for expected in expected_params]):
        # JSON params do not contain all expected params
        params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif not request.capability_info.lookup_capability_request(capability_request_id):
        # CapabilityRequest with given id does not exist; can't create a new version for it
        does_not_exist_msg = (f"Capability request #{capability_request_id} does not exist. Cannot create new version",)
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        new_capability_version = request.capability_info.create_capability_version(
            capability_request_id, parameters=params["parameters"]
        )
        return Response(json_body=new_capability_version.__json__())


@view_config(route_name="submit_capability_version", renderer="json")
def submit_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to submit a capability version for execution
    URL: capability/request/{capability_request_id}/version/{version_id}/submit

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_number = request.matchdict["version_number"]

    version = request.capability_info.lookup_version(capability_request_id, version_number)

    if not version_number:
        # Capability request not found
        does_not_exist_msg = (
            f"Capability request #{capability_request_id} v{version_number} does not exist; cannot submit it."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)

    execution = request.capability_launcher.launch_capability(version)
    return Response(json_body=execution.__json__())


@view_config(route_name="add_file_to_latest", renderer="json")
def add_file_to_latest(request: Request) -> Response:
    """
    Pyramid view that accepts a request to add a file to the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest/files/{filename}

    :param request: PUT request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])

    if capability_request:
        if capability_request.current_version:
            file = request.capability_info.save_version_file(
                capability_request.current_version, request.matchdict["filename"], request.body
            )
            return file
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_version_metadata", renderer="json")
def view_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        return version.workflow_metadata
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)


@view_config(route_name="edit_version_metadata", renderer="json")
def edit_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        version.workflow_metadata = request.json_body
        request.capability_info.save_version(version)
        return request.json_body
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)


@view_config(route_name="change_srdp_compatibility_version_metadata", renderer="json")
def change_srdp_compatibility_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to change the SRDP-Compatibility status of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/srdp_compatible

    :param request: GET request
    :return: 200 OK response with SRDP-Compatibility status of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
        or 500 response (HTTPInternalServerError) if request to archive is_srdp endpoint fails
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        settings = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings")
        archive_url = settings.locatorServiceUrlPrefix
        # split the locatorServiceUrlPrefix to get url without /location?
        archive_url = archive_url.split("/location?")
        version.parameters["metadata"]["is_srdp"] = request.json_body
        change_srdp_url = archive_url[0] + f"/is_srdp?locator={version.parameters['product_locator']}"
        r = requests.post(change_srdp_url, json=request.json_body)
        if r.status_code != requests.codes.ok:
            srdp_change_failed_msg = f"SRDP-Compatibility change failed: {r.status_code}"
            return HTTPInternalServerError(detail=srdp_change_failed_msg)
        request.capability_info.save_version(version)
        return request.json_body
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)
